unit UDrawText;

interface

type
	TFnBitBlt = procedure(xDest, yDest: integer; width, height: integer; xSour, ySour: integer; color: longword); stdcall;

procedure SmDrawText(xDest, yDest: integer; const szText: widestring; color: longword; fnBitBlt: TFnBitBlt);

implementation

type
	TGlyphRect = record
		nRect: integer;
		rect: array[0..2] of
		record
			x1, y1, x2, y2: integer;
		end;
	end;

const
	W_BUFFER: integer = 512;
	H_BUFFER: integer = 256;

//procedure ConvertUnicodeToJohap(const widecode: widechar; var johap: longword);
//end;

function  GlyphIndex(chCode : longword) : longint;
const
	MAX_SM1 = 19;
	MAX_SM2 = 21;
	MAX_SM3 = 28;

	UJIZ_TABLE: array[0..2, 0..31] of integer =
	(
		( 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1),
		( 3, 4, 5, 6, 7,10,11,12,13,14,15,18,19,20,21,22,23,26,27,28,29,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1),
		( 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,19,20,21,22,23,24,25,26,27,28,29,-1,-1,-1,-1)
	);
	INCOMPLETE_TABLE: array[$3131..$3163] of integer =
	(
		// CONSONANT
		$8800, $8C00, $8444, $9000, $8446, $8447, $9400, $9800, $9C00, $844A,
		$844B, $844C, $844D, $844E, $844F, $8450, $A000, $A400, $A800, $8454,
		$AC00, $B000, $B400, $B800, $BC00, $C000, $C400, $C800, $CC00, $D000,
		// VOWEL
		$8461, $8481, $84A1, $84C1, $84E1, $8541, $8561, $8581, $85A1, $85C1,
		$85E1, $8641, $8661, $8681, $86A1, $86C1, $86E1, $8741, $8761, $8781,
		$87A1
	);
	CONVERT_TABLE: array[0..2, 0..30] of byte =
	(
		( 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
		( 0, 0, 0, 1, 2, 3, 4, 5, 0, 0, 6, 7, 8, 9,10,11, 0, 0,12,13,14,15,16,17, 0, 0,18,19,20,21, 0),
		( 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16, 0,17,18,19,20,21,22,23,24,25,26,27, 0)
	);
var
	smgal_com : integer;
	SM1       : integer;
	SM2       : integer;
	SM3       : integer;
	temp16    : word;

	i         : integer;
	jamo      : array[0..2] of longword;
	johap     : longword;
begin
	johap := 0;
	smgal_com := integer(chCode);

	if (smgal_com < $0080) then begin
		johap := smgal_com;
	end
	else if ((smgal_com >= $AC00) and (smgal_com <= $D7A3)) then begin
		dec(smgal_com, $AC00);

		SM1 :=  smgal_com div (MAX_SM2 * MAX_SM3);
		SM2 := (smgal_com - SM1 * (MAX_SM2 * MAX_SM3)) div MAX_SM3;
		SM3 :=  smgal_com - SM1 * (MAX_SM2 * MAX_SM3) - SM2 * MAX_SM3;

		SM1 := UJIZ_TABLE[0][SM1];
		SM2 := UJIZ_TABLE[1][SM2];
		SM3 := UJIZ_TABLE[2][SM3];

		johap := $8000 or (SM1 shl 10) or (SM2 shl 5) or SM3;
	end
	else if ((smgal_com >= $3131) and (smgal_com <= $3163)) then begin
		temp16 := INCOMPLETE_TABLE[smgal_com];
		if temp16 > 0 then
			johap := temp16
	end;

	if ((johap and $8000) = 0) or (johap > $FFFF) then begin
		result := -1;
		exit;
	end;

	jamo[0] := (johap shr 10) and $1F;
	jamo[1] := (johap shr  5) and $1F;
	jamo[2] := (johap shr  0) and $1F;

	result := 0;
	for i := 0 to 2 do begin
		result := result shl 8;
		if jamo[i] < $20 then
			result := result or longint(CONVERT_TABLE[i, jamo[i]]);
	end;
end;

procedure GetGlyph(index: longword; out jamoIndex: array of integer);
var
	jamo: array[1..3] of longword;
	face: array[1..3] of longword;
begin
	jamo[1] := (index shr 16) and $000000FF;
	jamo[2] := (index shr  8) and $000000FF;
	jamo[3] := (index shr  0) and $000000FF;

	face[1] := 0;
	face[2] := 0;
	face[3] := 0;

	if jamo[3] > 0 then begin
		if jamo[2] in [1..8,21]            then face[1] := 95;
		if jamo[2] in [9,13,14,18,19]      then face[1] := 114;
		if jamo[2] in [10..12,15..17,20]   then face[1] := 133;
		if jamo[1] in [1,16]               then face[2] := 194;
		if jamo[1] in [2..15,17..19]       then face[2] := 215;
		if jamo[2] in [1,3,10]             then face[3] := 236;
		if jamo[2] in [5,7,12,15,17,20,21] then face[3] := 263;
		if jamo[2] in [2,4,6,8,11,16]      then face[3] := 290;
		if jamo[2] in [9,13,14,18,19]      then face[3] := 317;

		if (jamo[1] = 0) and (jamo[2] = 0) then
			face[3] := 236;

	end else begin
		if jamo[2] in [1..8,21]            then face[1] := 0;
		if jamo[2] in [9,13,19]            then face[1] := 19;
		if jamo[2] in [14,18]              then face[1] := 38;
		if jamo[2] in [10..12,20]          then face[1] := 57;
		if jamo[2] in [15..17]             then face[1] := 76;
		if jamo[1] in [0,1,16]             then face[2] := 152;
		if jamo[1] in [2..15,17..19]       then face[2] := 173;

		if jamo[1] = 0 then
			face[1] := 0;
	end;

	jamoIndex[0] := face[1] + jamo[1] - 1;

	if (jamo[2] > 0) then
		jamoIndex[1] := face[2] + jamo[2] - 1
	else
		jamoIndex[1] := -1;

	if ((jamo[3] > 0) and (face[3] > 0)) then
		jamoIndex[2] := face[3] + jamo[3] - 1
	else
		jamoIndex[2] := -1;
end;

function GetGlyphRect(index: longword; out glyphRect: TGlyphRect): boolean;
var
	i: integer;
	wExtent: integer;
	hExtent: integer;
	pitch: integer;
	jamoIndex: array[0..2] of integer;
begin
	if (integer(index) >= $0080) then begin
		wExtent := 16;
		hExtent := 16;
		pitch   := W_BUFFER div wExtent;

		GetGlyph(GlyphIndex(index), jamoIndex);

		glyphRect.nRect := 0;

		for i := 0 to 2 do begin
			if (jamoIndex[i] < 0) then
				break;

			index := W_BUFFER * 4 div wExtent + jamoIndex[i];

			glyphRect.rect[i].x1 := (index mod pitch) * wExtent;
			glyphRect.rect[i].y1 := (index div pitch) * hExtent;
			glyphRect.rect[i].x2 := glyphRect.rect[i].x1 + wExtent;
			glyphRect.rect[i].y2 := glyphRect.rect[i].y1 + hExtent;

			inc(glyphRect.nRect);
		end;
	end
	else begin
		wExtent := 8;
		hExtent := 16;
		pitch   := W_BUFFER div wExtent;

		index := W_BUFFER * 2 div wExtent + index;

		glyphRect.rect[0].x1 := (index mod pitch) * wExtent;
		glyphRect.rect[0].y1 := (index div pitch) * hExtent;
		glyphRect.rect[0].x2 := glyphRect.rect[0].x1 + wExtent;
		glyphRect.rect[0].y2 := glyphRect.rect[0].y1 + hExtent;

		glyphRect.nRect := 1;
	end;

	result := TRUE;
end;

procedure SmDrawText(xDest, yDest: integer; const szText: widestring; color: longword; fnBitBlt: TFnBitBlt);
var
	i, j, w, h: integer;
	glyphRect: TGlyphRect;
begin
	for j := 0 to pred(length(szText)) do begin
		GetGlyphRect(integer(szText[j+1]), glyphRect);

		w := glyphRect.rect[0].x2 - glyphRect.rect[0].x1;
		h := glyphRect.rect[0].y2 - glyphRect.rect[0].y1;

		for i := 0 to pred(glyphRect.nRect) do begin
			fnBitBlt(xDest, yDest, w, h, glyphRect.rect[i].x1, glyphRect.rect[i].y1, color);
		end;
		inc(xDest, glyphRect.rect[0].x2 - glyphRect.rect[0].x1);
	end;
end;

end.
