unit UDxApp;

interface

uses
	Windows, SysUtils, MMsystem, Direct3D9;

const
	SCREEN_WIDTH   = 800;
	SCREEN_HEIGHT  = 600;
	SCREEN_DEPTH   = 32;

	BITMAP_WIDTH   = 512;
	BITMAP_HEIGHT  = 256;

type
	TGameMain = class
	private
		m_hWindow        : THandle;
		m_pD3D           : IDirect3D9;
		m_pD3DDevice     : IDirect3DDevice9;
		m_pFontData      : IDirect3DTexture9;

	public
		procedure Init(window: THandle; width, height: integer; depth: integer; isFullScreen: boolean);
		procedure Done();
		procedure Run();

		function  BitBlt(xDest, yDest: integer; const hImage: IDirect3DTexture9; xSour, ySour, wSour, hSour: integer; color: longword = $FFFFFFFF): boolean;
		property  FontTexture: IDirect3DTexture9 read m_pFontData;
	end;

var
	g_GameMain : TGameMain;

implementation

uses
	UDrawText;

type
	PVertexTL = ^TVertexTL;
	TVertexTL = packed record
		x, y, z, rhw : single;
		color        : longword;
		tu, tv       : single;
	end;

// Callback Լ
procedure DxBitBlt(xDest, yDest: integer; width, height: integer; xSour, ySour: integer; color: longword); stdcall;
begin
	g_GameMain.BitBlt(xDest, yDest, g_GameMain.FontTexture, xSour, ySour, width, height, color);
end;

// TGameMain implementation
function TGameMain.BitBlt(xDest, yDest: integer; const hImage: IDirect3DTexture9; xSour, ySour, wSour, hSour: integer; color: longword): boolean;
const
	MAX_VERTEX = 4;
var
	x, y, w, h, tu, tv, tw, th: single;
	vertices: array[0..pred(MAX_VERTEX)] of TVertexTL;
begin
	result := FALSE;

	if not assigned(hImage) then
		exit;

	x  := xDest - 0.5;
	y  := yDest - 0.5;
	w  := wSour;
	h  := hSour;

	// ̷ ȵ... ̴ϱ
	tu := xSour / BITMAP_WIDTH;
	tv := ySour / BITMAP_HEIGHT;
	tw := wSour / BITMAP_WIDTH;
	th := hSour / BITMAP_HEIGHT;

	// ȭ ؾ 
	vertices[0].x := x;
	vertices[0].y := y;
	vertices[0].z := 1.0;
	vertices[0].rhw := 1.0;
	vertices[0].tu := tu;
	vertices[0].tv := tv;

	vertices[1].x := x + w;
	vertices[1].y := y;
	vertices[1].z := 1.0;
	vertices[1].rhw := 1.0;
	vertices[1].tu := tu + tw;
	vertices[1].tv := tv;

	vertices[2].x := x + w;
	vertices[2].y := y + h;
	vertices[2].z := 1.0;
	vertices[2].rhw := 1.0;
	vertices[2].tu := tu + tw;
	vertices[2].tv := tv + th;

	vertices[3].x := x;
	vertices[3].y := y + h;
	vertices[3].z := 1.0;
	vertices[3].rhw := 1.0;
	vertices[3].tu := tu;
	vertices[3].tv := tv + th;

	vertices[0].color := color;
	vertices[1].color := color;
	vertices[2].color := color;
	vertices[3].color := color;

	m_pD3DDevice.SetTexture(0, hImage);
	m_pD3DDevice.SetFVF(D3DFVF_XYZRHW or D3DFVF_DIFFUSE or D3DFVF_TEX1);
	m_pD3DDevice.DrawPrimitiveUP(D3DPT_TRIANGLEFAN, MAX_VERTEX-2, @vertices, sizeof(TVertexTL));

	result := TRUE;
end;

procedure TGameMain.Init(window: THandle; width, height: integer; depth: integer; isFullScreen: boolean);
	procedure s_AdjustWindowRect(hWindow: THandle; width, height: integer);
	var
		winInfo: TWindowInfo;
		rect: TRect;
	begin
		if not Windows.GetWindowInfo(hWindow, winInfo) then
			exit;

		rect.Left   := 0;
		rect.Top    := 0;
		rect.Right  := width;
		rect.Bottom := height;

		if not Windows.AdjustWindowRect(rect, winInfo.dwStyle, (GetMenu(hWindow) <> 0)) then
			exit;

		Windows.MoveWindow(hWindow, winInfo.rcWindow.Left, winInfo.rcWindow.Top, rect.Right - rect.Left, rect.Bottom - rect.Top, TRUE);
	end;

var
	d3dPP: TD3DPresentParameters;
	lockRect: TD3DLockedRect;
	x, y: integer;
	pf: file;
	srcBuffer: array[0.. pred(BITMAP_WIDTH div 8)] of byte;
	pDest32: PDWORD;

begin
	m_hWindow    := window;
	m_pD3D       := nil;
	m_pD3DDevice := nil;
	m_pFontData  := nil;

	// Direct3D9 νϽ 
	m_pD3D := Direct3DCreate9(D3D_SDK_VERSION);
	assert(Assigned(m_pD3D));

	if not isFullScreen then
		s_AdjustWindowRect(window, width, height);

	// Device 
	ZeroMemory(@d3dPP, sizeof(d3dPP));

	d3dPP.SwapEffect       := D3DSWAPEFFECT_DISCARD;
	d3dPP.BackBufferCount  := 1;
	d3dPP.Flags            := D3DPRESENTFLAG_LOCKABLE_BACKBUFFER;
	d3dPP.Windowed         := not isFullScreen;
	d3dPP.BackBufferFormat := D3DFMT_X8R8G8B8;
	d3dPP.BackBufferWidth  := width;
	d3dPP.BackBufferHeight := height;

	m_pD3D.CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, window, D3DCREATE_SOFTWARE_VERTEXPROCESSING, d3dPP, m_pD3DDevice);
	assert(Assigned(m_pD3DDevice));

	//  ÿ Alpha test ʿϴ
	m_pD3DDevice.SetRenderState(D3DRS_ALPHATESTENABLE, 1);
	m_pD3DDevice.SetRenderState(D3DRS_ALPHAREF, $00);
	m_pD3DDevice.SetRenderState(D3DRS_ALPHAFUNC, D3DCMP_NOTEQUAL);

	// BMP ҷñ
	m_pD3DDevice.CreateTexture(BITMAP_WIDTH, BITMAP_HEIGHT, 0, 0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, m_pFontData, nil);
	assert(Assigned(m_pFontData));

	if m_pFontData.LockRect(0, lockRect, nil, 0) = D3D_OK then begin
		AssignFile(pf, 'a1bit.bmp');
		Reset(pf, 1);
		// ̴
		Seek(pf, FileSize(pf) - (BITMAP_WIDTH * BITMAP_HEIGHT div 8));

		for y := 0 to pred(BITMAP_HEIGHT) do begin
			BlockRead(pf, srcBuffer, sizeof(srcBuffer));
			pDest32 := PDWORD(longword(lockRect.Bits) + (pred(BITMAP_HEIGHT)-y) * lockRect.Pitch);
			for x := 0 to pred(BITMAP_WIDTH) do begin
				if (srcBuffer[x div 8] and ($80 shr (x mod 8))) > 0 then
					pDest32^ := $FFFFFFFF
				else
					pDest32^ := $00000000;
				inc(pDest32);
			end;
		end;

		CloseFile(pf);

		m_pFontData.UnlockRect(0);
	end;
end;

procedure TGameMain.Done();
begin
	m_pFontData  := nil;
	m_pD3DDevice := nil;
	m_pD3D       := nil;
end;

procedure TGameMain.Run();
const
	predTick: longint = 0;
var
	i: integer;
	currTick: longint;
begin
	m_pD3DDevice.Clear(0, nil, D3DCLEAR_TARGET, $FFFFFFFF, 1.0, 0);
	m_pD3DDevice.BeginScene();

	begin
		//  , ̹  ٽ
		for i := 1 to pred(SCREEN_HEIGHT div 16) do begin
			SmDrawText(0, i*16, 'Ƹٿ  ABCDEFGhijklmn!@#$%^&*()1234567890  ڶ ±ر տ  Ƹٿ  ABC', Random($FFFFFFFF), DxBitBlt);
		end;

		// FPS üũ
		if (predTick > 0) then begin
			currTick := timeGetTime() - predTick;
			if currTick > 0 then begin
				SmDrawText(0, 0, Format('%d', [1000 div currTick]), $FF2040FF, DxBitBlt);
			end;
		end;
		predTick := timeGetTime();
	end;

	m_pD3DDevice.EndScene();
	m_pD3DDevice.Present(nil, nil, m_hWindow, nil);
end;

end.
